/*
 * Copyright (c) 1980 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * Modified for the Linux SCSI tape driver by Kai Makisara
 * 
 * 25 Sep 94, Eyal Lebedinsky (eyal@ise.canberra.edu.au) [0.5.1eyal]
 * 	- add more options to 'drvbuffer'.
 * 	- allow 'k' as suffix to sizes.
 * 	- add tape type SCSI2.
 * 29 Apr 96, Eyal Lebedinsky (eyal@ise.canberra.edu.au) [1.5.2eyal]
 * 	- upgrade to new ioctls (partitions).
 * 26 Jan 97, Eyal Lebedinsky (eyal@eyal.emu.id.au) [0.5.3eyal]
 * 	- add 'compress' operation.
 * 	- update the man page.
 * 
*/

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1980 The Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)mt.c	5.6 (Berkeley) 6/6/91";
#endif /* not lint */

/*
 * mt --
 *   magnetic tape manipulation program
 */
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <linux/mtio.h>


static struct commands {
	char *c_name;
	int c_code;
	int c_ronly;
} com[] = {
	{ "weof",	MTWEOF,	0 },
	{ "eof",	MTWEOF,	0 },
	{ "fsf",	MTFSF,	1 },
	{ "fsfm",	MTFSFM,	1 },
	{ "bsf",	MTBSF,	1 },
	{ "bsfm",	MTBSFM,	1 },
	{ "fsr",	MTFSR,	1 },
	{ "bsr",	MTBSR,	1 },
	{ "rewind",	MTREW,	1 },
	{ "offline",	MTOFFL,	1 },
	{ "rewoffl",	MTOFFL,	1 },
	{ "retension",	MTRETEN,1 },
	{ "eod",	MTEOM,	1 },
	{ "seod",	MTEOM,	1 },
	{ "seek",	MTSEEK, 1 },
	{ "tell",	MTTELL,	1 },
	{ "status",	MTNOP,	1 },
	{ "erase",	MTERASE, 0},
	{ "setblk",	MTSETBLK, 1},
	{ "setdensity",	MTSETDENSITY, 1},
	{ "drvbuffer",	MTSETDRVBUFFER, 1},
	{ "setpart",	MTSETPART, 1 },
	{ "mkpart",	MTMKPART, 1 },
	{ "compress",	MTCOMPRESSION, 1 },
	{ 0 }
};

static int mtfd;
static struct mtop mt_com;
static struct mtget mt_status;
static struct mtpos mt_pos;
static char *tape;

static void status (struct mtget *bp);
static void printreg (char *s, unsigned short v, unsigned char *bits);


static void
usage (void)
{
	struct commands *comp;

	fprintf(stderr, "usage: mt [-h] [-v] [-f device] command [count]\n");
	fprintf(stderr, "\tcommand is:\n");
	for (comp = com; comp->c_name != NULL; comp++)
		fprintf (stderr, "\t\t%s\n", comp->c_name);
}

int
main (int argc, char **argv)
{
	char	*getenv();
	struct commands *comp;
	int	part;
	char	*command;
	char	*count;
	int	nargs;
	char	*p;
	int	noerr;

	if ((tape = getenv("TAPE")) == NULL)
		tape = DEFTAPE;
	command = NULL;
	count = NULL;
	noerr = 0;
	for (nargs = 0; argc > 1; ++argv, --argc) {
		if ('-' != argv[1][0]) {
			switch (++nargs) {
			case 1:
				command = argv[1];
				break;
			case 2:
				count = argv[1];
				break;
			default:
				usage ();
				exit (1);
			}
			continue;
		}
		switch (argv[1][1]) {
		case 't':
		case 'f':
			tape = argv[2];
			--argc;
			++argv;
			break;
		case 'h':
			usage ();
			noerr = 1;
			break;
		case 'v':
			fprintf (stderr, "mt version 0.5.3eyal\n");
			noerr = 1;
			break;
		default:
			usage ();
			exit (1);
			break;
		}
	}
	if (NULL == command) {
		if (!noerr)
			usage ();
		exit(1);
	}
	for (comp = com; comp->c_name != NULL; comp++)
		if (strncmp(command, comp->c_name, strlen(command)) == 0)
			break;
	if (comp->c_name == NULL) {
		fprintf(stderr, "mt: don't grok \"%s\"\n", command);
		exit(1);
	}
	if ((mtfd = open(tape, comp->c_ronly ? O_RDONLY : O_RDWR)) < 0) {
		perror(tape);
		exit(1);
	}
	if (comp->c_code == MTTELL) {
		if (ioctl(mtfd, MTIOCPOS, (char *)&mt_pos) < 0) {
			perror("mt");
			exit(2);
		}
		if (ioctl(mtfd, MTIOCGET, (char *)&mt_status) < 0) {
			perror("mt");
			exit(2);
		}
		if (0 != (part = mt_status.mt_resid & 0x0ff))
			printf("%d+", part);
		printf("%ld\n", mt_pos.mt_blkno);
	}
	else if (comp->c_code != MTNOP) {
		mt_com.mt_op = comp->c_code;
		if (NULL == count) {
			if (MTSETPART == comp->c_code ||
					MTMKPART == comp->c_code) {
				printf("Missing numeric argument\n");
				exit (1);
			}
			mt_com.mt_count = 1;
		}
		else if (MTSETDRVBUFFER == comp->c_code) {
			if ('b' == count[0]) {
				p = count + 1;
				if ('+' == *p) {
					mt_com.mt_count = MT_ST_SETBOOLEANS;
					++p;
				} else if ('-' == *p) {
					mt_com.mt_count = MT_ST_CLEARBOOLEANS;
					++p;
				} else
					mt_com.mt_count = MT_ST_BOOLEANS;
				for (; *p; ++p) {
				    switch (*p) {
				    case 'w':
					mt_com.mt_count |= MT_ST_BUFFER_WRITES;
					break;
				    case 'a':
					mt_com.mt_count |= MT_ST_ASYNC_WRITES;
					break;
				    case 'r':
					mt_com.mt_count |= MT_ST_READ_AHEAD;
					break;
				    case 'd':
					mt_com.mt_count |= MT_ST_DEBUGGING;
					break;
				    case 't':
					mt_com.mt_count |= MT_ST_TWO_FM;
					break;
				    case 'f':
					mt_com.mt_count |= MT_ST_FAST_MTEOM;
					break;
				    case 'l':
					mt_com.mt_count |= MT_ST_AUTO_LOCK;
					break;
				    case 'b':
					mt_com.mt_count |= MT_ST_CAN_BSR;
					break;
				    case 'p':
					mt_com.mt_count |= MT_ST_CAN_PARTITIONS;
					break;
				    case '2':
					mt_com.mt_count |= MT_ST_SCSI2LOGICAL;
					break;
				    default:
					fprintf(stderr, 
						"mt: bad boolean option\n");
					exit(1);
					break;
				    }
				}
			} else if ('t' == count[0]) {
				mt_com.mt_count = atoi(count+1);
				if ('k' == count[strlen(count)-1])
					mt_com.mt_count *= 1024;
				mt_com.mt_count |= MT_ST_WRITE_THRESHOLD;
			} else if (count[0] < '0' || count[0] > '7' ||
							count[1]) {
				fprintf(stderr, "mt: bad buffer mode\n");
				exit(1);
			} else
				mt_com.mt_count = count[0] - '0';
		} else {
			if (MTSEEK == comp->c_code &&
				NULL != (p = strchr (count, '+'))) {
				char	*t;

				if (NULL == (t = strdup (count))) {
					fprintf (stderr, "mt: strdup failed\n");
					exit (2);
				}
				t[p - count] = '\0';
				mt_com.mt_count = atoi (t);
				free (t);
				if (mt_com.mt_count & ~0x0ff) {
					fprintf (stderr, "mt: bad prtition\n");
					exit (1);
				}
				mt_com.mt_op = MTSETPART;
				if (ioctl(mtfd, MTIOCTOP, (char *)&mt_com) < 0) {
					fprintf(stderr, "%s %s %d ", 
						tape, comp->c_name,
						mt_com.mt_count);
					perror("failed");
					exit(2);
				}
				mt_com.mt_op = MTSEEK;
				mt_com.mt_count = atoi(p+1);
			} else
				mt_com.mt_count = atoi(count);
			if ('k' == count[strlen(count)-1])
				mt_com.mt_count *= 1024;
		}
		if (mt_com.mt_count < 0) {
			fprintf(stderr, "mt: negative repeat count\n");
			exit(1);
		}
		if (MTSETPART == comp->c_code && mt_com.mt_count > 0x0ff) {
			fprintf (stderr, "mt: partition > 255\n");
			exit (1);
		}
		if (ioctl(mtfd, MTIOCTOP, (char *)&mt_com) < 0) {
			fprintf(stderr, "%s %s %d ", tape, comp->c_name,
				mt_com.mt_count);
			perror("failed");
			exit(2);
		}
	} else {
		if (ioctl(mtfd, MTIOCGET, (char *)&mt_status) < 0) {
			perror("mt");
			exit(2);
		}
		status(&mt_status);
	}
	return 0;
}

#ifdef vax
#include <vaxmba/mtreg.h>
#include <vaxmba/htreg.h>

#include <vaxuba/utreg.h>
#include <vaxuba/tmreg.h>
#undef b_repcnt		/* argh */
#include <vaxuba/tsreg.h>
#endif

#ifdef sun
#include <sundev/tmreg.h>
#include <sundev/arreg.h>
#endif

#ifdef tahoe
#include <tahoe/vba/cyreg.h>
#endif

static struct tape_desc {
	short	t_type;		/* type of magtape device */
	char	*t_name;	/* printing name */
	unsigned char	*t_dsbits;	/* "drive status" register */
	unsigned char	*t_erbits;	/* "error" register */
} tapes[] = {
#ifdef vax
	{ MT_ISTS,	"ts11",		0,		TSXS0_BITS },
	{ MT_ISHT,	"tm03",		HTDS_BITS,	HTER_BITS },
	{ MT_ISTM,	"tm11",		0,		TMER_BITS },
	{ MT_ISMT,	"tu78",		MTDS_BITS,	0 },
	{ MT_ISUT,	"tu45",		UTDS_BITS,	UTER_BITS },
#endif
#ifdef sun
	{ MT_ISCPC,	"TapeMaster",	TMS_BITS,	0 },
	{ MT_ISAR,	"Archive",	ARCH_CTRL_BITS,	ARCH_BITS },
#endif
#ifdef tahoe
	{ MT_ISCY,	"cipher",	CYS_BITS,	CYCW_BITS },
#endif
#ifdef linux
	{ MT_ISSCSI1,	"SCSI 1",	0,		0 },
	{ MT_ISSCSI2,	"SCSI 2",	0,		0 },
#endif
	{ 0 }
};

/*
 * Interpret the status buffer returned
 */
static void
status (struct mtget *bp)
{
	struct tape_desc *mt;

	for (mt = tapes; mt->t_type; mt++)
		if (mt->t_type == bp->mt_type)
			break;
	if (mt->t_type == 0) {
		printf("unknown tape drive type (%ld)\n", bp->mt_type);
		return;
	}
	printf ("%s tape drive, residual=%ld\n", mt->t_name, bp->mt_resid);
	printreg ("ds", bp->mt_dsreg, mt->t_dsbits);
	printreg ("\ner", bp->mt_erreg, mt->t_erbits);
	putchar ('\n');
}

/*
 * Print a register a la the %b format of the kernel's printf
 */
static void
printreg (char *s, unsigned short v, unsigned char *bits)
{
	int i, any = 0;
	int c;

	if (bits && *bits == 8)
		printf("%s=%o", s, v);
	else
		printf("%s=%x", s, v);
	bits++;
	if (v && bits) {
		putchar('<');
		while (0 != (i = *bits++)) {
			if (v & (1 << (i-1))) {
				if (any)
					putchar(',');
				any = 1;
				for (; (c = *bits) > 32; bits++)
					putchar(c);
			} else
				for (; *bits > 32; bits++)
					;
		}
		putchar('>');
	}
}
